package com.dust.tripartite.netmc;

import com.dust.tripartite.netmc.codec.DelimiterBasedFrameDecoder;
import com.dust.tripartite.netmc.codec.LengthFieldAndDelimiterFrameDecoder;
import com.dust.tripartite.netmc.codec.MessageDecoderWrapper;
import com.dust.tripartite.netmc.codec.MessageEncoderWrapper;
import io.netty.bootstrap.ServerBootstrap;
import io.netty.channel.ChannelFuture;
import io.netty.channel.ChannelInitializer;
import io.netty.channel.EventLoopGroup;
import io.netty.channel.nio.NioEventLoopGroup;
import io.netty.channel.socket.nio.NioChannelOption;
import io.netty.channel.socket.nio.NioServerSocketChannel;
import io.netty.channel.socket.nio.NioSocketChannel;
import io.netty.handler.codec.ByteToMessageDecoder;
import io.netty.handler.timeout.IdleStateHandler;
import io.netty.util.NettyRuntime;
import io.netty.util.concurrent.Future;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;

/**
 * @author yezhihao
 * home https://gitee.com/yezhihao/jt808-server
 */
public class TCPServer {

    private static final Logger log = LoggerFactory.getLogger(TCPServer.class);
    private final AtomicBoolean isRunning = new AtomicBoolean(false);

    private EventLoopGroup bossGroup = null;
    private EventLoopGroup workerGroup = null;

    private final String name;
    private final NettyConfig config;

    public TCPServer(String name, NettyConfig config) {
        this.name = name;
        this.config = config;
    }

    private void startInternal() {
        try {
            this.bossGroup = new NioEventLoopGroup(1);
            this.workerGroup = new NioEventLoopGroup(NettyRuntime.availableProcessors());
            ServerBootstrap bootstrap = new ServerBootstrap();
            bootstrap.channel(NioServerSocketChannel.class);
            bootstrap.group(bossGroup, workerGroup);
            bootstrap.option(NioChannelOption.SO_BACKLOG, 1024)
                    .option(NioChannelOption.SO_REUSEADDR, true)
                    .childOption(NioChannelOption.TCP_NODELAY, true)
                    .childHandler(new ChannelInitializer<NioSocketChannel>() {

                        private final MessageEncoderWrapper messageEncoderWrapper = new MessageEncoderWrapper(config.encoder);
                        private final MessageDecoderWrapper messageDecoderWrapper = new MessageDecoderWrapper(config.decoder);

                        @Override
                        public void initChannel(NioSocketChannel channel) {
                            channel.pipeline()
                                    .addLast(new IdleStateHandler(config.readerIdleTime, config.writerIdleTime, config.allIdleTime, TimeUnit.SECONDS))
                                    .addLast("frameDecoder", frameDecoder())
                                    .addLast("decoder", messageDecoderWrapper)
                                    .addLast("encoder", messageEncoderWrapper)
                                    .addLast("adapter", config.adapter);
                        }
                    });

            ChannelFuture channelFuture = bootstrap.bind(config.port).sync();
            log.warn("==={}启动成功, port={}===", name, config.port);
            channelFuture.channel().closeFuture().sync();
        } catch (Exception e) {
            log.warn("==={}出现异常, port={}===", e, config.port);
        } finally {
            stop();
        }
    }

    public ByteToMessageDecoder frameDecoder() {
        if (config.lengthField == null)
            return new DelimiterBasedFrameDecoder(config.maxFrameLength, config.delimiter);
        return new LengthFieldAndDelimiterFrameDecoder(config.maxFrameLength, config.lengthField, config.delimiter);
    }

    public void start() {
        if (!isRunning.compareAndSet(false, true)) {
            log.warn("==={}已经启动, port={}===", name, config.port);
            return;
        }
        new Thread(this::startInternal).start();
    }

    public void stop() {
        if (!isRunning.compareAndSet(true, false)) {
            log.warn("==={}已经停止, port={}===", name, config.port);
            return;
        }
        Future<?> future = this.bossGroup.shutdownGracefully();
        if (!future.isSuccess()) {
            log.warn("bossGroup 无法正常停止", future.cause());
        }
        future = this.workerGroup.shutdownGracefully();
        if (!future.isSuccess()) {
            log.warn("workerGroup 无法正常停止", future.cause());
        }
        log.warn("==={}已经停止, port={}===", name, config.port);
    }
}